# Copyright Red Hat
#
# snapm/manager/plugins/lvm2.py - Snapshot Manager LVM2 plugins
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: Apache-2.0
"""
LVM2 snapshot manager plugins
"""
from subprocess import run, CalledProcessError
from json import loads, JSONDecodeError
from math import floor
from os.path import join as path_join
from os import stat, environ
from stat import S_ISBLK
from time import time
from shutil import which

from snapm import (
    SnapmInvalidIdentifierError,
    SnapmCalloutError,
    SnapmBusyError,
    SnapmNoSpaceError,
    SnapmNotFoundError,
    SnapmPluginError,
    SizePolicy,
    SnapStatus,
    Snapshot,
)
from snapm.manager.plugins import (
    DEV_PREFIX,
    DEV_MAPPER_PREFIX,
    DMSETUP_CMD,
    DMSETUP_INFO,
    DMSETUP_NO_HEADINGS,
    DMSETUP_COLUMNS,
    DMSETUP_FIELDS_UUID,
    Plugin,
    parse_snapshot_name,
    device_from_mount_point,
    mount_point_space_used,
    format_snapshot_name,
    encode_mount_point,
)

#: Maximum length for LVM2 LV names
LVM_MAX_NAME_LEN = 127
#: Length of extension for LVM2 CoW snapshot LV name
LVM_COW_SNAPSHOT_NAME_LEN = 4

# Global LVM2 report options
LVM_REPORT_FORMAT = "--reportformat"
LVM_JSON = "json"
LVM_JSON_STD = "json_std"
LVM_BASIC = "basic"
LVM_OPTIONS = "--options"
LVM_UNITS = "--units"
LVM_BYTES = "b"
LVM_UUID_PREFIX = "LVM-"

# Main lvm executable
LVM_CMD = "lvm"

# Version subcommand
LVM_VERSION = "version"

# LVM version string prefix
LVM_VERSION_STR = "LVM version"

# lvs report options
LVS_CMD = "lvs"
LVS_REPORT = "report"
LVS_ALL = "--all"
LVS_LV = "lv"
LVS_LV_NAME = "lv_name"
LVS_VG_NAME = "vg_name"
LVS_LV_ATTR = "lv_attr"
LVS_LV_ROLE = "lv_role"
LVS_LV_ORIGIN = "origin"
LVS_POOL_LV = "pool_lv"
LVS_LV_SIZE = "lv_size"
LVS_DATA_PERCENT = "data_percent"
LVS_LV_HIDDEN_START = "["
LVS_LV_HIDDEN_END = "]"
LVS_FIELD_OPTIONS = (
    "vg_name,lv_name,lv_attr,origin,pool_lv,lv_size,data_percent,lv_role"
)

# lvs options for devpath to vg/lv name
LVS_FIELD_MIN_OPTIONS = "vg_name,lv_name"
LVS_NO_HEADINGS = "--noheadings"


# lv_attr flag values
LVM_COW_SNAP_ATTR = "s"
LVM_MERGE_SNAP_ATTR = "S"
LVM_THIN_VOL_ATTR = "V"
LVM_COW_ORIGIN_ATTR = "o"
LVM_ACTIVE_ATTR = "a"
LVM_INVALID_ATTR = "I"
LVM_LV_TYPE_DEFAULT = "-"
LVM_LV_ORIGIN_MERGING = "O"
LVM_SKIP_ACTIVATION_ATTR = "k"

# lv_attr flag indexes
LVM_LV_STATE_ATTR_IDX = 4
LVM_LV_SKIP_ACTIVATION_IDX = 9

# lv_role strings
LVM_COW_SNAPSHOT_ROLE = "thicksnapshot"
LVM_THIN_SNAPSHOT_ROLE = "thinsnapshot"

# vgs report options
VGS_CMD = "vgs"
VGS_REPORT = "report"
VGS_VG = "vg"
VGS_FIELD_OPTIONS = "vg_name,vg_free"

# lvcreate command options
LVCREATE_CMD = "lvcreate"
LVCREATE_SNAPSHOT = "--snapshot"
LVCREATE_NAME = "--name"
LVCREATE_SIZE = "--size"

# lvremove command options
LVREMOVE_CMD = "lvremove"
LVREMOVE_YES = "--yes"

# lvrename command
LVRENAME_CMD = "lvrename"

# lvchange command
LVCHANGE_CMD = "lvchange"

# lvchange command options
LVCHANGE_ACTIVATE = "--activate"
LVCHANGE_ACTIVE_YES = "y"
LVCHANGE_ACTIVE_NO = "n"
LVCHANGE_YES = "--yes"
LVCHANGE_IGNOREACTIVATIONSKIP = "--ignoreactivationskip"
LVCHANGE_SETACTIVATIONSKIP = "--setactivationskip"
LVCHANGE_ACTIVATIONSKIP_YES = "y"
LVCHANGE_ACTIVATIONSKIP_NO = "n"

# lvconvert command
LVCONVERT_CMD = "lvconvert"

# lvconvert command options
LVCONVERT_MERGE = "--merge"

# lvresize command
LVRESIZE_CMD = "lvresize"

# Minimum possible LVM2 CoW snapshot size (512MiB)
MIN_LVM2_COW_SNAPSHOT_SIZE = 512 * 1024**2

# Maximum time to cache lvs data in seconds
LVS_CACHE_VALID = 5

# LVM commands required by the plugin
_LVM_CMDS = [
    DMSETUP_CMD,
    LVM_CMD,
    LVS_CMD,
    VGS_CMD,
    LVCREATE_CMD,
    LVREMOVE_CMD,
    LVRENAME_CMD,
    LVCHANGE_CMD,
    LVCONVERT_CMD,
]

MINIMUM_LVM_VERSION = (2, 3, 11)
MINIMUM_LVM_VERSION_JSON_STD = (2, 3, 17)

# LVM2 environment variables to filter out
_LVM_ENV_FILTER = [
    "LVM_OUT_FD",
    "LVM_ERR_FD",
    "LVM_REPORT_FD",
    "LVM_COMMAND_PROFILE",
    "LVM_RUN_BY_DMEVENTD",
    "LVM_SUPPRESS_FD_WARNINGS",
    "LVM_SUPPRESS_SYSLOG",
    "LVM_VG_NAME",
    "LVM_LVMPOLLD_PIDFILE",
    "LVM_LVMPOLLD_SOCKET",
    "LVM_LOG_FILE_EPOCH",
    "LVM_LOG_FILE_MAX_LINES",
    "LVM_EXPECTED_EXIT_STATUS",
    "LVM_SUPPRESS_LOCKING_FAILURE_MESSAGES",
    "DM_ABORT_ON_INTERNAL_ERRORS",
    "DM_DISABLE_UDEV",
    "DM_DEBUG_WITH_LINE_NUMBERS",
]


def _decode_stderr(err):
    """
    Decode and strip the stderr member of a ``CalledProcessError`` and
    return the result as a string.

    :param err: A ``CalledProcessError`` like exception.
    :returns: A stripped string representation of the exception's stderr
              member.
    """
    return err.stderr.decode("utf8").strip()


def _check_lvm_present():
    """
    Check for the presence of the required LVM2 commands.

    :returns: ``True`` if all commands are present and executable or
              ``False`` otherwise.
    """
    if not all(which(cmd) for cmd in _LVM_CMDS):
        raise SnapmNotFoundError("LVM2 commands not found")


def _get_lvm_version():
    """
    Return the installed version of LVM2 as a tuple.

    :returns: A version tuple (major, minor, patch) of LVM2
    """

    def _version_string_to_tuple(version):
        return tuple(map(int, version.split(".")))

    lvm_cmd_args = [LVM_CMD, LVM_VERSION]
    lvm_cmd = run(lvm_cmd_args, capture_output=True, check=True)
    lvm_version_info = str.strip(lvm_cmd.stdout.decode("utf8"))
    for line in lvm_version_info.splitlines():
        if LVM_VERSION_STR in line:
            (_, _, version, _) = line.split()
            if "(" in version:
                version, _ = version.split("(", maxsplit=1)
            return _version_string_to_tuple(version)
    return (0, 0, 0)


def is_lvm_device(devpath):
    """
    Test whether ``devpath`` is an LVM device.

    Return ``True`` if the device at ``devpath`` is an LVM device or
    ``False`` otherwise.
    """

    if devpath.startswith(DEV_MAPPER_PREFIX):
        dm_name = devpath.removeprefix(DEV_MAPPER_PREFIX)
    else:
        dm_name = devpath

    dmsetup_cmd_args = [
        DMSETUP_CMD,
        DMSETUP_INFO,
        DMSETUP_COLUMNS,
        DMSETUP_NO_HEADINGS,
        DMSETUP_FIELDS_UUID,
        dm_name,
    ]
    try:
        dmsetup_cmd = run(dmsetup_cmd_args, capture_output=True, check=True)
    except CalledProcessError as err:
        raise SnapmCalloutError(
            f"Error calling {DMSETUP_CMD}: {_decode_stderr(err)}"
        ) from err
    uuid = dmsetup_cmd.stdout.decode("utf8").strip()
    return uuid.startswith(LVM_UUID_PREFIX)


def vg_lv_from_origin(origin):
    """
    Return a ``(vg_name, lv_name)`` tuple for the LVM device with origin
    path ``origin``.
    """
    name_parts = origin.removeprefix(DEV_PREFIX + "/").split("/")
    return (name_parts[0], name_parts[1])


class Lvm2Snapshot(Snapshot):
    """
    Class for LVM2 snapshot objects.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name,
        snapset_name,
        origin,
        timestamp,
        mount_point,
        provider,
        vg_name,
        lv_name,
        lv_dict=None,
    ):
        super().__init__(name, snapset_name, origin, timestamp, mount_point, provider)
        self.vg_name = vg_name
        self.lv_name = lv_name
        if lv_dict:
            self._lv_dict_cache = lv_dict
        else:
            lvs_dict = self.provider.get_lvs_json_report(
                f"{self.vg_name}/{self.lv_name}"
            )
            self._lv_dict_cache = lvs_dict[LVS_REPORT][0][LVS_LV][0]
        self._lv_dict_cache_ts = time()

    def __str__(self):
        return "".join(
            [
                super().__str__(),
                f"\nVolumeGroup:    {self.vg_name}",
                f"\nLogicalVolume:  {self.lv_name}",
            ]
        )

    @property
    def origin(self):
        return path_join(DEV_PREFIX, self.vg_name, self._origin)

    @property
    def origin_options(self):
        """
        File system options needed to specify the origin of this snapshot.

        No file system options are requires for LVM2 block snapshots: always
        return the empty string.
        """
        return ""

    def _devpath(self):
        return path_join(DEV_PREFIX, self.vg_name, self.lv_name)

    @property
    def devpath(self):
        if self.status != SnapStatus.ACTIVE:
            return ""
        return self._devpath()

    @property
    def status(self):
        lv_dict = self._get_lv_dict_cache()
        lv_attr = lv_dict[LVS_LV_ATTR]
        if lv_attr[LVM_LV_STATE_ATTR_IDX] == LVM_INVALID_ATTR:
            return SnapStatus.INVALID
        if lv_attr.startswith(LVM_MERGE_SNAP_ATTR):
            return SnapStatus.REVERTING
        if lv_attr[LVM_LV_STATE_ATTR_IDX] == LVM_ACTIVE_ATTR:
            return SnapStatus.ACTIVE
        return SnapStatus.INACTIVE

    @property
    def size(self):
        lv_dict = self._get_lv_dict_cache()
        return int(lv_dict[LVS_LV_SIZE].rstrip("B"))

    @property
    def free(self):
        raise NotImplementedError

    # Pylint does not understand the decorator notation.
    # pylint: disable=invalid-overridden-method
    @Snapshot.autoactivate.getter
    def autoactivate(self):
        lv_dict = self._get_lv_dict_cache()
        lv_attr = lv_dict[LVS_LV_ATTR]
        if lv_attr[LVM_LV_SKIP_ACTIVATION_IDX] == LVM_SKIP_ACTIVATION_ATTR:
            return False
        return True

    def invalidate_cache(self):
        self._lv_dict_cache = None
        self._lv_dict_cache_ts = 0

    def _get_lv_dict_cache(self):
        now = time()
        if not self._lv_dict_cache or (self._lv_dict_cache_ts + LVS_CACHE_VALID) < now:
            lvs_dict = self.provider.get_lvs_json_report(
                f"{self.vg_name}/{self.lv_name}"
            )
            self._lv_dict_cache = lvs_dict[LVS_REPORT][0][LVS_LV][0]
            self._lv_dict_cache_ts = now
        return self._lv_dict_cache


class Lvm2CowSnapshot(Lvm2Snapshot):
    """
    Class for LVM2 copy-on-write snapshot objects.
    """

    @property
    def free(self):
        lv_dict = self._get_lv_dict_cache()
        lv_data_percent = float(lv_dict[LVS_DATA_PERCENT])
        return int(((100.0 - lv_data_percent) * self.size) / 100.0)


class Lvm2ThinSnapshot(Lvm2Snapshot):
    """
    Class for LVM2 thin snapshot objects.
    """

    @property
    def free(self):
        lv_dict = self._get_lv_dict_cache()
        return self.provider.pool_free_space(self.vg_name, lv_dict[LVS_POOL_LV])


def _volume_type_or_merging(lv_dict, lv_type):
    """
    Return ``True`` if the logical volume represented by ``lv_dict`` is either
    the specified ``lv_type``, or a merging snapshot volume, or ``False``
    otherwise.
    """
    if lv_dict[LVS_LV_ATTR].startswith(lv_type):
        return True
    if lv_dict[LVS_LV_ATTR].startswith(LVM_MERGE_SNAP_ATTR):
        return True
    return False


def filter_cow_snapshot(lv_dict):
    """
    Filter LVM2 CoW snapshots.

    Return ``True`` if the logical volume represented by ``lv_dict`` is an LVM2
    COW snapshot or ``False`` otherwise. The ``lv_dict`` argument must be a
    dictionary representing the output of the ``lvs`` reporting command.
    """
    if not _volume_type_or_merging(lv_dict, LVM_COW_SNAP_ATTR):
        return False
    if LVM_COW_SNAPSHOT_ROLE not in lv_dict[LVS_LV_ROLE]:
        return False
    if lv_dict[LVS_LV_ORIGIN] == "":
        return False
    return True


def filter_thin_snapshot(lv_dict):
    """
    Filter LVM2 thin snapshots.

    Return ``True`` if the logical volume represented by ``lv_dict`` is an LVM2
    thin snapshot or ``False`` otherwise. The ``lv_dict`` argument must be a
    dictionary representing the output of the ``lvs`` reporting command.
    """
    if not _volume_type_or_merging(lv_dict, LVM_THIN_VOL_ATTR):
        return False
    if LVM_THIN_SNAPSHOT_ROLE not in lv_dict[LVS_LV_ROLE]:
        return False
    if lv_dict[LVS_LV_ORIGIN] == "":
        return False
    return True


def _activate(active, name, silent=False):
    """
    Call lvchange to activate or deactivate an LVM2 volume.

    :param name: The name of the LV to operate on.
    :param silent: ``True`` if errors should not be propagated or
                   ``False`` otherwise.
    """
    lvchange_cmd = [
        LVCHANGE_CMD,
        LVCHANGE_YES,
        LVCHANGE_IGNOREACTIVATIONSKIP,
        LVCHANGE_ACTIVATE,
        active,
        name,
    ]
    try:
        run(lvchange_cmd, capture_output=True, check=True)
    except CalledProcessError as err:
        if not silent:
            raise SnapmCalloutError(
                f"{LVCHANGE_CMD} failed with: {_decode_stderr(err)}"
            ) from err


class _Lvm2(Plugin):
    """
    Abstract base class for LVM2 snapshot plugins.
    """

    max_name_len = LVM_MAX_NAME_LEN

    def vg_lv_from_device_path(self, devpath):
        """
        Return a ``(vg_name, lv_name)`` tuple for the LVM device at
        ``devpath``.
        """
        lvs_cmd_args = [
            LVS_CMD,
            LVS_NO_HEADINGS,
            LVM_REPORT_FORMAT,
            LVM_BASIC,
            LVM_OPTIONS,
            LVS_FIELD_MIN_OPTIONS,
            devpath,
        ]
        try:
            lvs_cmd = run(lvs_cmd_args, env=self._env, capture_output=True, check=True)
        except CalledProcessError as err:
            raise SnapmCalloutError(
                f"Error calling {LVS_CMD}: {_decode_stderr(err)}"
            ) from err
        name = lvs_cmd.stdout.decode("utf8").strip()
        name_parts = name.split(" ")
        return (name_parts[0], name_parts[1])

    def pool_name_from_vg_lv(self, vg_lv):
        """
        Return the thin pool associated with the logical volume identified by
        ``vg_lv``.
        """
        lvs_dict = self.get_lvs_json_report(vg_lv)
        lv_dict = lvs_dict[LVS_REPORT][0][LVS_LV][0]
        return lv_dict[LVS_POOL_LV]

    def vg_free_space(self, vg_name):
        """
        Return the free space available as bytes for the volume group named
        ``vg_name``.
        """
        vgs_dict = self.get_vgs_json_report(vg_name=vg_name)
        for vg_dict in vgs_dict[VGS_REPORT][0][VGS_VG]:
            if vg_dict["vg_name"] == vg_name:
                return int(vg_dict["vg_free"].rstrip("B"))
        raise ValueError(f"Volume group {vg_name} not found")

    def lv_dev_size(self, vg_name, lv_name):
        """
        Return the size of the specified logical volume in bytes.
        """
        lvs_dict = self.get_lvs_json_report(f"{vg_name}/{lv_name}")
        for lv_dict in lvs_dict[LVS_REPORT][0][LVS_LV]:
            if lv_dict["vg_name"] == vg_name and lv_dict["lv_name"] == lv_name:
                return int(lv_dict["lv_size"].rstrip("B"))
        raise ValueError(f"Logical volume {vg_name}/{lv_name} not found")

    def pool_free_space(self, vg_name, pool_name):
        """
        Return the free space available as bytes for the thin pool identified
        by ``vg_name`` and ``pool_name``.
        """
        lvs_dict = self.get_lvs_json_report(f"{vg_name}/{pool_name}")
        lv_dict = lvs_dict[LVS_REPORT][0][LVS_LV][0]
        data_percent = float(lv_dict[LVS_DATA_PERCENT])
        pool_size = int(lv_dict[LVS_LV_SIZE].rstrip("B"))
        return int(pool_size - floor((pool_size * data_percent) / 100.0))

    def get_lvs_json_report(self, vg_lv=None, lvs_all=False):
        """
        Call out to the ``lvs`` program and return a report in JSON format.
        """
        lvs_cmd_args = [
            LVS_CMD,
            LVM_REPORT_FORMAT,
            self._json_fmt,
            LVM_UNITS,
            LVM_BYTES,
            LVM_OPTIONS,
            LVS_FIELD_OPTIONS,
        ]
        if vg_lv:
            lvs_cmd_args.append(vg_lv)
        if lvs_all:
            lvs_cmd_args.append(LVS_ALL)
        try:
            lvs_cmd = run(lvs_cmd_args, capture_output=True, check=True)
        except CalledProcessError as err:
            raise SnapmCalloutError(
                f"Error calling {LVS_CMD}: {_decode_stderr(err)}"
            ) from err
        try:
            lvs_dict = loads(lvs_cmd.stdout)
        except JSONDecodeError as err:
            raise SnapmCalloutError(
                f"Unable to decode {LVS_CMD} JSON output: {err}"
            ) from err
        return lvs_dict

    def get_vgs_json_report(self, vg_name=None):
        """
        Call out to the ``vgs`` program and return a report in JSON format.
        """
        vgs_cmd_args = [
            VGS_CMD,
            LVM_REPORT_FORMAT,
            self._json_fmt,
            LVM_UNITS,
            LVM_BYTES,
            LVM_OPTIONS,
            VGS_FIELD_OPTIONS,
        ]
        if vg_name:
            vgs_cmd_args.append(vg_name)
        try:
            vgs_cmd = run(vgs_cmd_args, capture_output=True, check=True)
        except CalledProcessError as err:
            raise SnapmCalloutError(
                f"Error calling {VGS_CMD}: {_decode_stderr(err)}"
            ) from err
        try:
            vgs_dict = loads(vgs_cmd.stdout)
        except JSONDecodeError as err:
            raise SnapmCalloutError(
                f"Unable to decode {VGS_CMD} JSON output: {err}"
            ) from err
        return vgs_dict

    def _check_lvm_version(self):
        """
        Check for the required minimum LVM2 version and enable any version
        dependent workarounds as needed.
        """

        def _version_string(value):
            return f"{value[0]}.{value[1]}.{value[2]}"

        try:
            lvm_version = _get_lvm_version()
        except CalledProcessError as err:
            raise SnapmPluginError(
                f"Error getting LVM2 version: {_decode_stderr(err)}"
            ) from err
        if lvm_version < MINIMUM_LVM_VERSION:
            raise SnapmPluginError(
                f"Unsupported LVM2 version: {_version_string(lvm_version)} "
                f"< {_version_string(MINIMUM_LVM_VERSION)}"
            )

        # Work around missing --reportformat json_std in ubuntu-24.04
        # See https://github.com/snapshotmanager/snapm/issues/110
        if lvm_version < MINIMUM_LVM_VERSION_JSON_STD:
            self._json_fmt = LVM_JSON

    def _sanitize_environment(self):
        env = environ.copy()
        for var in _LVM_ENV_FILTER:
            if var in env:
                env.pop(var)
        return env

    def __init__(self, logger):
        super().__init__(logger)

        # Default to using json_std when available.
        self._json_fmt = LVM_JSON_STD

        # Sanitize environment for LVM2 callouts.
        self._env = self._sanitize_environment()

        # Check for presence of required LVM2 binaries.
        _check_lvm_present()

        # Check LVM2 minimum version requirements.
        self._check_lvm_version()

    def discover_snapshots(self):
        """
        Discover snapshots managed by this plugin class.

        Returns a list of objects that are a subclass of ``Snapshot``.
        """
        raise NotImplementedError

    def can_snapshot(self, source):
        """
        Test whether this plugin can snapshot the specified block device or
        mount point path.

        :param source: The mount point or block device path to test.
        :returns: ``True`` if this plugin can snapshot the file system or
                  block device at ``source``, or ``False`` otherwise.
        """
        raise NotImplementedError

    # pylint: disable=too-many-arguments
    def check_create_snapshot(
        self, origin, snapset_name, timestamp, mount_point, size_policy
    ):
        """
        Perform pre-creation checks before creating a snapshot.

        :param origin: The origin volume for the snapshot.
        :param snapset_name: The name of the snapshot set to be created.
        :param timestamp: The snapshot set timestamp.
        :param mount_point: The mount point path for this snapshot.
        :raises: ``SnapmNoSpaceError`` if there is insufficient free space to
                 create the snapshot.
        """
        raise NotImplementedError

    # pylint: disable=too-many-arguments
    def create_snapshot(
        self, origin, snapset_name, timestamp, mount_point, size_policy
    ):
        """
        Create a snapshot of ``origin`` in the snapset named ``snapset_name``.

        :param origin: The origin volume for the snapshot.
        :param snapset_name: The name of the snapshot set to be created.
        :param timestamp: The snapshot set timestamp.
        :param mount_point: The mount point path for this snapshot.
        :raises: ``SnapmNoSpaceError`` if there is insufficient free space to
                 create the snapshot.
        """
        raise NotImplementedError

    def origin_from_mount_point(self, mount_point):
        """
        Return a string representing the origin from a given mount point path.
        """
        if S_ISBLK(stat(mount_point).st_mode):
            device = mount_point
        else:
            device = device_from_mount_point(mount_point)
        if not is_lvm_device(device):
            return None
        (vg_name, lv_name) = self.vg_lv_from_device_path(device)
        return path_join(DEV_PREFIX, vg_name, lv_name)

    def delete_snapshot(self, name):
        """
        Delete the snapshot named ``name``

        :param name: The name of the snapshot to be removed.
        """
        lvremove_cmd = [LVREMOVE_CMD, LVREMOVE_YES, name]
        try:
            run(lvremove_cmd, env=self._env, capture_output=True, check=True)
        except CalledProcessError as err:
            raise SnapmCalloutError(
                f"{LVREMOVE_CMD} failed with: {_decode_stderr(err)}"
            ) from err

    # pylint: disable=too-many-arguments
    def rename_snapshot(self, old_name, origin, snapset_name, timestamp, mount_point):
        """
        Rename the snapshot named ``old_name`` according to the provided
        snapshot field values.

        :param old_name: The original name of the snapshot to be renamed.
        :param origin: The origin volume for the snapshot.
        :param snapset_name: The new name of the snapshot set.
        :param timestamp: The snapshot set timestamp.
        :param mount_point: The mount point of the snapshot.
        """
        vg_name, lv_name = vg_lv_from_origin(origin)
        new_name = format_snapshot_name(
            lv_name, snapset_name, timestamp, encode_mount_point(mount_point)
        )

        self._log_debug(
            "Renaming snapshot from %s to %s",
            old_name,
            new_name,
        )

        lvrename_cmd = [
            LVRENAME_CMD,
            vg_name,
            old_name,
            new_name,
        ]
        try:
            run(lvrename_cmd, env=self._env, capture_output=True, check=True)
        except CalledProcessError as err:
            raise SnapmCalloutError(
                f"{LVRENAME_CMD} failed with: {_decode_stderr(err)}"
            ) from err

        return self._build_snapshot(
            f"{vg_name}/{new_name}",
            snapset_name,
            lv_name,
            timestamp,
            mount_point,
            self,
            vg_name,
            new_name,
        )

    def check_revert_snapshot(self, name, origin):
        """
        Check whether this snapshot can be reverted or not. This method returns
        if the current snapshot can be reverted and raises an exception if not.

        :returns: None
        :raises: ``NotImplementedError`` if this plugin does not support the
        revert operation, ``SnapmBusyError`` if the snapshot is already in the
        process of being reverted to another snapshot state or
        ``SnapmPluginError`` if another reason prevents the snapshot from being
        merged.
        """
        vg_name, lv_name = vg_lv_from_origin(origin)
        lvs_dict = self.get_lvs_json_report(f"{vg_name}/{lv_name}")
        lv_report = lvs_dict[LVS_REPORT][0][LVS_LV][0]
        lv_attr = lv_report[LVS_LV_ATTR]
        if lv_attr[0] == LVM_LV_ORIGIN_MERGING:
            raise SnapmBusyError(
                f"Snapshot revert is in progress for {name} origin volume {vg_name}/{lv_name}"
            )

    def revert_snapshot(self, name):
        """
        Revert the state of the content of the origin to the content at the
        time the snapshot was taken.

        For LVM2 snapshots of in-use logical volumes this will take place at
        the next activation (typically a reboot into the revert boot entry
        for the snapshot set).
        """
        lvconvert_cmd = [
            LVCONVERT_CMD,
            LVCONVERT_MERGE,
            name,
        ]
        try:
            run(lvconvert_cmd, env=self._env, capture_output=False, check=True)
        except CalledProcessError as err:
            raise SnapmCalloutError(
                f"{LVCONVERT_CMD} failed with: {_decode_stderr(err)}"
            ) from err

    def activate_snapshot(self, name):
        """
        Activate the snapshot named ``name``

        :param name: The name of the snapshot to be activated.
        """
        self._log_debug("Activating %s snapshot %s", self.name, name)
        _activate(LVCHANGE_ACTIVE_YES, name)

    def deactivate_snapshot(self, name):
        """
        Deactivate the snapshot named ``name``

        :param name: The name of the snapshot to be deactivated.
        """
        self._log_debug("Deactivating %s snapshot %s", self.name, name)
        _activate(LVCHANGE_ACTIVE_NO, name, silent=True)

    def set_autoactivate(self, name, auto=False):
        """
        Set the autoactivation state of the snapshot named ``name``.

        :param name: The name of the snapshot to be modified.
        :param auto: ``True`` to enable autoactivation or ``False`` otherwise.
        """
        lvchange_cmd = [
            LVCHANGE_CMD,
            LVCHANGE_SETACTIVATIONSKIP,
            LVCHANGE_ACTIVATIONSKIP_NO if auto else LVCHANGE_ACTIVATIONSKIP_YES,
            name,
        ]
        try:
            run(lvchange_cmd, env=self._env, capture_output=True, check=True)
        except CalledProcessError as err:
            raise SnapmCalloutError(
                f"{LVCHANGE_CMD} failed with: {_decode_stderr(err)}"
            ) from err

    def _check_lvm_name(self, vg_name, lv_name):
        """
        Check whether a proposed LVM name exceeds the maximum allowed name
        length for an LVM2 logical volume.

        :param vg_name: The volume group name to check.
        :param lv_name: The logical volume name to check.
        :raises: ``SnapmInvalidIdentifierError`` if the proposed name exceeds
                 limits.
        """
        full_name = f"{vg_name}/{lv_name}"
        if len(full_name) > self.max_name_len:
            raise SnapmInvalidIdentifierError(
                f"Logical volume name {full_name} exceeds maximum LVM2 name length"
            )

    def _build_snapshot(
        self,
        name,
        snapset_name,
        origin,
        timestamp,
        mount_point,
        provider,
        vg_name,
        lv_name,
        lv_dict=None,
    ):
        """
        Build an instance of this plugin's snapshot class and return it.
        """
        raise NotImplementedError


def _snapshot_min_size(space_used):
    """
    Return the minimum snapshot size given the space used by the snapshot
    mount point.

    :param space_used: The space currently in use on the mount point.
    :returns: The greater of ``space_used`` and ``MIN_LVM2_COW_SNAPSHOT_SIZE``.
    """
    return max(MIN_LVM2_COW_SNAPSHOT_SIZE, space_used)


class Lvm2Cow(_Lvm2):
    """
    LVM2 Copy-on-write snapshot plugin.
    """

    name = "lvm2-cow"
    version = "0.1.0"
    snapshot_class = Lvm2CowSnapshot

    max_name_len = LVM_MAX_NAME_LEN - LVM_COW_SNAPSHOT_NAME_LEN

    def discover_snapshots(self):
        """
        Discover snapshots managed by the lvm2-cow plugin.

        :returns: A list of ``Lvm2Snapshot`` objects discovered by this plugin.
        """
        snapshots = []
        lvs_dict = self.get_lvs_json_report(lvs_all=True)
        for lv_dict in lvs_dict[LVS_REPORT][0][LVS_LV]:
            if filter_cow_snapshot(lv_dict):
                try:
                    lv_name = lv_dict[LVS_LV_NAME]
                    if lv_name.startswith(LVS_LV_HIDDEN_START):
                        lv_name = lv_name[1:-1]
                    fields = parse_snapshot_name(lv_name, lv_dict[LVS_LV_ORIGIN])
                except ValueError:
                    continue
                if fields is not None:
                    (snapset, timestamp, mount_point) = fields
                    full_name = f"{lv_dict[LVS_VG_NAME]}/{lv_name}"
                    self._log_debug("Found %s snapshot: %s", self.name, full_name)
                    snapshots.append(
                        Lvm2CowSnapshot(
                            full_name,
                            snapset,
                            f"{lv_dict[LVS_LV_ORIGIN]}",
                            timestamp,
                            mount_point,
                            self,
                            f"{lv_dict[LVS_VG_NAME]}",
                            f"{lv_name}",
                            lv_dict=lv_dict,
                        )
                    )
        return snapshots

    def can_snapshot(self, source):
        """
        Test whether the lvm2-cow plugin can snapshot the specified ``source``.

        :param source: The mount point or block device path to test.
        :returns: ``True`` if this plugin can snapshot the file system or block
                  device found at ``source``, or ``False`` otherwise.
        """

        if S_ISBLK(stat(source).st_mode):
            device = source
        else:
            device = device_from_mount_point(source)

        if not is_lvm_device(device):
            return False
        (vg_name, lv_name) = self.vg_lv_from_device_path(device)
        lvs_dict = self.get_lvs_json_report(f"{vg_name}/{lv_name}")
        lv_report = lvs_dict[LVS_REPORT][0][LVS_LV][0]
        lv_attr = lv_report[LVS_LV_ATTR]
        if lv_attr[0] == LVM_LV_ORIGIN_MERGING:
            raise SnapmBusyError(
                f"Snapshot revert is in progress for {self.name} origin volume {vg_name}/{lv_name}"
            )
        if lv_attr[0] != LVM_LV_TYPE_DEFAULT and lv_attr[0] != LVM_COW_ORIGIN_ATTR:
            return False
        return True

    def _check_free_space(self, origin, mount_point, size_policy):
        """
        Check for available space in volume group ``vg_name`` for the specified
        mount point.

        :param vg_name: The name of the volume group to check.
        :param mount_point: The mount point path to check.
        :returns: The space used on the mount point.
        :raises: ``SnapmNoSpaceError`` if the minimum snapshot size exceeds the
                 available space.
        """
        vg_name, lv_name = vg_lv_from_origin(origin)
        fs_used = mount_point_space_used(mount_point)
        vg_free = self.vg_free_space(vg_name)
        lv_size = self.lv_dev_size(vg_name, lv_name)
        policy = SizePolicy(origin, mount_point, vg_free, fs_used, lv_size, size_policy)
        snapshot_min_size = _snapshot_min_size(policy.size)
        if vg_free < (sum(self.size_map[vg_name].values()) + snapshot_min_size):
            raise SnapmNoSpaceError(
                f"Volume group {vg_name} has insufficient free space to snapshot {mount_point}"
            )
        return snapshot_min_size

    def check_create_snapshot(
        self, origin, snapset_name, timestamp, mount_point, size_policy
    ):
        vg_name, lv_name = vg_lv_from_origin(origin)
        snapshot_name = format_snapshot_name(
            lv_name, snapset_name, timestamp, encode_mount_point(mount_point)
        )
        self._check_lvm_name(vg_name, snapshot_name)
        if vg_name not in self.size_map:
            self.size_map[vg_name] = {}
        self.size_map[vg_name][lv_name] = self._check_free_space(
            origin, mount_point, size_policy
        )

    def create_snapshot(
        self, origin, snapset_name, timestamp, mount_point, size_policy
    ):
        vg_name, lv_name = vg_lv_from_origin(origin)
        self._log_debug(
            "Creating CoW snapshot for %s/%s %s %s",
            vg_name,
            lv_name,
            "mounted at" if mount_point else "",
            mount_point,
        )
        snapshot_name = format_snapshot_name(
            lv_name, snapset_name, timestamp, encode_mount_point(mount_point)
        )
        self._check_lvm_name(vg_name, snapshot_name)
        snapshot_size = self.size_map[vg_name][lv_name]
        lvcreate_cmd = [
            LVCREATE_CMD,
            LVCREATE_SNAPSHOT,
            LVCREATE_NAME,
            snapshot_name,
            LVCREATE_SIZE,
            f"{snapshot_size}b",
            origin,
        ]
        try:
            run(lvcreate_cmd, env=self._env, capture_output=True, check=True)
        except CalledProcessError as err:
            raise SnapmCalloutError(
                f"{LVCREATE_CMD} failed with: {_decode_stderr(err)}"
            ) from err
        return Lvm2CowSnapshot(
            f"{vg_name}/{snapshot_name}",
            snapset_name,
            lv_name,
            timestamp,
            mount_point,
            self,
            vg_name,
            snapshot_name,
        )

    def check_resize_snapshot(self, name, origin, mount_point, size_policy):
        vg_name, lv_name = vg_lv_from_origin(origin)
        if vg_name not in self.size_map:
            self.size_map[vg_name] = {}
        self.size_map[vg_name][lv_name] = self._check_free_space(
            origin, mount_point, size_policy
        )

    def resize_snapshot(self, name, origin, mount_point, size_policy):
        vg_name, lv_name = vg_lv_from_origin(origin)
        snapshot_size = self.size_map[vg_name][lv_name]

        lvresize_cmd = [
            LVRESIZE_CMD,
            LVCREATE_SIZE,
            f"{snapshot_size}b",
            name,
        ]
        try:
            run(lvresize_cmd, env=self._env, capture_output=True, check=True)
        except CalledProcessError as err:
            raise SnapmCalloutError(
                f"{LVRESIZE_CMD} failed with: {_decode_stderr(err)}"
            ) from err

    def _build_snapshot(
        self,
        name,
        snapset_name,
        origin,
        timestamp,
        mount_point,
        provider,
        vg_name,
        lv_name,
        lv_dict=None,
    ):
        """
        Build an instance of `Lvm2CowSnapshot` and return it.
        """
        return Lvm2CowSnapshot(
            name,
            snapset_name,
            origin,
            timestamp,
            mount_point,
            provider,
            vg_name,
            lv_name,
            lv_dict=lv_dict,
        )


class Lvm2Thin(_Lvm2):
    """
    LVM2 thin provisioning snapshot plugin.
    """

    name = "lvm2-thin"
    version = "0.1.0"
    snapshot_class = Lvm2ThinSnapshot

    max_name_len = LVM_MAX_NAME_LEN

    def discover_snapshots(self):
        snapshots = []
        lvs_dict = self.get_lvs_json_report(lvs_all=True)
        for lv_dict in lvs_dict[LVS_REPORT][0][LVS_LV]:
            if filter_thin_snapshot(lv_dict):
                try:
                    lv_name = lv_dict[LVS_LV_NAME]
                    if lv_name.startswith(LVS_LV_HIDDEN_START):
                        lv_name = lv_name[1:-1]
                    fields = parse_snapshot_name(lv_name, lv_dict[LVS_LV_ORIGIN])
                except ValueError:
                    continue
                if fields is not None:
                    full_name = f"{lv_dict[LVS_VG_NAME]}/{lv_name}"
                    self._log_debug("Found %s snapshot: %s", self.name, full_name)
                    (snapset, timestamp, mount_point) = fields
                    snapshots.append(
                        Lvm2ThinSnapshot(
                            full_name,
                            snapset,
                            f"{lv_dict[LVS_LV_ORIGIN]}",
                            timestamp,
                            mount_point,
                            self,
                            f"{lv_dict[LVS_VG_NAME]}",
                            f"{lv_name}",
                            lv_dict=lv_dict,
                        )
                    )
        return snapshots

    def can_snapshot(self, source):
        """
        Test whether the lvm2-thin plugin can snapshot the specified ``source``.

        :param source: The mount point or block device path to test.
        :returns: ``True`` if this plugin can snapshot the file system or block
                  device found at ``source``, or ``False`` otherwise.
        """
        if S_ISBLK(stat(source).st_mode):
            device = source
        else:
            device = device_from_mount_point(source)

        if not is_lvm_device(device):
            return False
        (vg_name, lv_name) = self.vg_lv_from_device_path(device)
        lvs_dict = self.get_lvs_json_report(f"{vg_name}/{lv_name}")
        lv_report = lvs_dict[LVS_REPORT][0][LVS_LV][0]
        lv_attr = lv_report[LVS_LV_ATTR]
        if lv_attr[0] == LVM_LV_ORIGIN_MERGING:
            raise SnapmBusyError(
                f"Snapshot revert is in progress for {self.name} origin volume {vg_name}/{lv_name}"
            )
        if lv_attr[0] != LVM_THIN_VOL_ATTR:
            return False
        return True

    def _check_free_space(self, origin, pool_name, mount_point, size_policy):
        vg_name, lv_name = vg_lv_from_origin(origin)
        fs_used = mount_point_space_used(mount_point)
        lv_size = self.lv_dev_size(vg_name, lv_name)
        pool_free = self.pool_free_space(vg_name, pool_name)
        policy = SizePolicy(
            origin, mount_point, pool_free, fs_used, lv_size, size_policy
        )
        snapshot_min_size = policy.size
        if pool_free < (
            sum(self.size_map[vg_name][pool_name].values()) + snapshot_min_size
        ):
            raise SnapmNoSpaceError(
                f"Volume group thin pool {vg_name}/{pool_name} "
                f"has insufficient free space to snapshot {mount_point}"
            )
        return snapshot_min_size

    def check_create_snapshot(
        self, origin, snapset_name, timestamp, mount_point, size_policy
    ):
        vg_name, lv_name = vg_lv_from_origin(origin)
        pool_name = self.pool_name_from_vg_lv(origin)
        snapshot_name = format_snapshot_name(
            lv_name, snapset_name, timestamp, encode_mount_point(mount_point)
        )
        self._check_lvm_name(vg_name, snapshot_name)
        if vg_name not in self.size_map:
            self.size_map[vg_name] = {}
        if pool_name not in self.size_map[vg_name]:
            self.size_map[vg_name][pool_name] = {}
        self.size_map[vg_name][pool_name][lv_name] = self._check_free_space(
            origin, pool_name, mount_point, size_policy
        )

    def create_snapshot(
        self, origin, snapset_name, timestamp, mount_point, size_policy
    ):
        vg_name, lv_name = vg_lv_from_origin(origin)
        self._log_debug(
            "Creating thin snapshot for %s/%s mounted at %s",
            vg_name,
            lv_name,
            mount_point,
        )
        snapshot_name = format_snapshot_name(
            lv_name, snapset_name, timestamp, encode_mount_point(mount_point)
        )
        self._check_lvm_name(vg_name, snapshot_name)

        lvcreate_cmd = [
            LVCREATE_CMD,
            LVCREATE_SNAPSHOT,
            LVCREATE_NAME,
            snapshot_name,
            origin,
        ]
        try:
            run(lvcreate_cmd, env=self._env, capture_output=True, check=True)
        except CalledProcessError as err:
            raise SnapmCalloutError(
                f"{LVCREATE_CMD} failed with: {_decode_stderr(err)}"
            ) from err
        return Lvm2ThinSnapshot(
            f"{vg_name}/{snapshot_name}",
            snapset_name,
            lv_name,
            timestamp,
            mount_point,
            self,
            vg_name,
            snapshot_name,
        )

    def check_resize_snapshot(self, name, origin, mount_point, size_policy):
        vg_name, lv_name = vg_lv_from_origin(origin)
        pool_name = self.pool_name_from_vg_lv(origin)
        if vg_name not in self.size_map:
            self.size_map[vg_name] = {}
        if pool_name not in self.size_map[vg_name]:
            self.size_map[vg_name][pool_name] = {}
        self.size_map[vg_name][pool_name][lv_name] = self._check_free_space(
            origin, pool_name, mount_point, size_policy
        )

    def resize_snapshot(self, name, origin, mount_point, size_policy):
        pass

    def _build_snapshot(
        self,
        name,
        snapset_name,
        origin,
        timestamp,
        mount_point,
        provider,
        vg_name,
        lv_name,
        lv_dict=None,
    ):
        """
        Build an instance of `Lvm2CowSnapshot` and return it.
        """
        return Lvm2ThinSnapshot(
            name,
            snapset_name,
            origin,
            timestamp,
            mount_point,
            provider,
            vg_name,
            lv_name,
            lv_dict=lv_dict,
        )



# Copyright 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Check version of stratisd
"""

# isort: THIRDPARTY
from packaging.specifiers import SpecifierSet, Version

from ._errors import StratisCliStratisdVersionError
from ._connection import get_object
from ._constants import MAXIMUM_STRATISD_VERSION, MINIMUM_STRATISD_VERSION, TOP_OBJECT


def check_stratisd_version():
    """
    Checks that the version of stratisd that is running is compatible with
    this version of the CLI.

    :raises StratisCliStratisdVersionError
    """
    # pylint: disable=import-outside-toplevel
    from ._data import Manager0

    version_spec = SpecifierSet(f">={MINIMUM_STRATISD_VERSION}") & SpecifierSet(
        f"<{MAXIMUM_STRATISD_VERSION}"
    )
    version = Manager0.Properties.Version.Get(get_object(TOP_OBJECT))

    if Version(version) not in version_spec:
        raise StratisCliStratisdVersionError(
            version, MINIMUM_STRATISD_VERSION, MAXIMUM_STRATISD_VERSION
        )



# Copyright Red Hat
#
# snapm/manager/_manager.py - Snapshot Manager
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: Apache-2.0
"""
Manager interface and plugin infrastructure.
"""
from subprocess import run, CalledProcessError
import logging
from time import time
from math import floor
from stat import S_ISBLK
from os import stat
from os.path import exists, ismount, normpath, samefile

from snapm.manager.signals import suspend_signals
from snapm.manager.boot import (
    BootCache,
    create_snapset_boot_entry,
    delete_snapset_boot_entry,
    create_snapset_revert_entry,
    delete_snapset_revert_entry,
    check_boom_config,
)

from snapm import (
    SNAPM_DEBUG_MANAGER,
    SNAPM_VALID_NAME_CHARS,
    SnapmError,
    SnapmCalloutError,
    SnapmNoSpaceError,
    SnapmNoProviderError,
    SnapmExistsError,
    SnapmBusyError,
    SnapmPathError,
    SnapmNotFoundError,
    SnapmInvalidIdentifierError,
    SnapmPluginError,
    SnapmStateError,
    SnapmRecursionError,
    SnapmArgumentError,
    Selection,
    bool_to_yes_no,
    is_size_policy,
    SnapStatus,
    SnapshotSet,
)
from ._loader import load_plugins


_log = logging.getLogger(__name__)
_log.set_debug_mask(SNAPM_DEBUG_MANAGER)

_log_debug = _log.debug
_log_debug_manager = _log.debug_masked
_log_info = _log.info
_log_warn = _log.warning
_log_error = _log.error

JOURNALCTL_CMD = "journalctl"


# pylint: disable=too-many-return-statements
def select_snapshot_set(select, snapshot_set):
    """
    Test SnapshotSet against Selection criteria.

    Test the supplied ``SnapshotSet`` against the selection criteria
    in ``s`` and return ``True`` if it passes, or ``False``
    otherwise.

    :param s: The selection criteria
    :param be: The SnapshotSet to test
    :rtype: bool
    :returns: True if SnapshotSet passes selection or ``False``
              otherwise.
    """
    if select.name and select.name != snapshot_set.name:
        return False
    if select.uuid and select.uuid != snapshot_set.uuid:
        return False
    if select.timestamp and select.timestamp != snapshot_set.timestamp:
        return False
    if select.nr_snapshots and select.nr_snapshots != snapshot_set.nr_snapshots:
        return False
    if select.mount_points:
        s_mount_points = sorted(select.mount_points)
        mount_points = sorted(snapshot_set.mount_points)
        if s_mount_points != mount_points:
            return False

    return True


def select_snapshot(select, snapshot):
    """
    Test SnapshotSet against Selection criteria.

    Test the supplied ``SnapshotSet`` against the selection criteria
    in ``s`` and return ``True`` if it passes, or ``False``
    otherwise.

    :param s: The selection criteria
    :param be: The SnapshotSet to test
    :rtype: bool
    :returns: True if SnapshotSet passes selection or ``False``
              otherwise.
    """
    if not select_snapshot_set(select, snapshot.snapshot_set):
        return False
    if select.snapshot_uuid and select.snapshot_uuid != snapshot.uuid:
        return False
    if select.snapshot_name and select.snapshot_name != snapshot.name:
        return False

    return True


def _suspend_journal():
    """
    Suspend journal writes to /var before creating snapshots.
    """
    try:
        run([JOURNALCTL_CMD, "--flush"], check=True)
        run([JOURNALCTL_CMD, "--relinquish-var"], check=True)
    except CalledProcessError as err:  # pragma: no cover
        raise SnapmCalloutError(
            f"Error calling journalctl to flush journal: {err}"
        ) from err


def _resume_journal():
    """
    Resume journal writes to /var after creating snapshots.
    """
    try:
        run([JOURNALCTL_CMD, "--flush"], check=True)
    except CalledProcessError as err:  # pragma: no cover
        raise SnapmCalloutError(
            f"Error calling journalctl to flush journal: {err}"
        ) from err


def _parse_source_specs(source_specs, default_size_policy):
    """
    Parse and normalize source paths and size policies.

    :param source_specs: A list of mount point or block device paths and
                         optional size policy strings.
    :returns: A tuple (sources, size_policies) containing a list of
              normalized mount point and block device paths and a dictionary
              mapping source paths to size policies.
    """
    sources = []
    size_policies = {}

    # Parse size policies and normalise mount paths
    for spec in source_specs:
        if ":" in spec:
            (source, policy) = spec.rsplit(":", maxsplit=1)
            if not is_size_policy(policy):
                source = f"{source}:{policy}"
                policy = default_size_policy
        else:
            (source, policy) = (spec, default_size_policy)
        source = normpath(source)
        sources.append(source)
        size_policies[source] = policy
    return (sources, size_policies)


def _check_revert_snapshot_set(snapset):
    """
    Check that all snapshots in a snapshot set can be reverted.
    Returns if all checks pass or raises an exception otherwise.

    :returns: None
    :raises: ``NotImplementedError``, ``SnapmBusyError`` or
             ``SnapmPluginError``
    """
    if snapset.status == SnapStatus.INVALID:
        raise SnapmStateError(
            f"Cannot revert snapset {snapset.name} with invalid snapshots"
        )

    # Check for ability to revert all snapshots in set
    for snapshot in snapset.snapshots:
        try:
            snapshot.check_revert()
        except NotImplementedError as err:
            _log_error(
                "Snapshot provider %s does not support revert",
                snapshot.provider.name,
            )
            raise SnapmPluginError from err
        except SnapmBusyError as err:
            _log_error(
                "Revert already in progress for snapshot origin %s",
                snapshot.origin,
            )
            raise SnapmBusyError(
                f"A revert has already started for snapshot set {snapset.name}"
            ) from err
        except SnapmPluginError as err:
            _log_error(
                "Revert prechecks failed for snapshot %s (%s)",
                snapshot.name,
                snapshot.provider.name,
            )
            raise err


def _check_snapset_status(snapset, operation):
    """
    Check that a snapshot set status is not ``SnapStatus.INVALID`` or
    ``SnapStatus.REVERTING`` before carrying out ``operation``.
    """
    if snapset.status in (SnapStatus.INVALID, SnapStatus.REVERTING):
        if snapset.status == SnapStatus.INVALID:
            _log_error("Cannot operate on invalid snapshot set '%s'", snapset.name)
        if snapset.status == SnapStatus.REVERTING:
            _log_error("Cannot operate on reverting snapshot set '%s'", snapset.name)
        raise SnapmStateError(
            f"Failed to {operation} snapset '{snapset.name}': status is '{snapset.status}'"
        )


def _find_mount_point_for_devpath(devpath):
    """
    Return the first mount point found in /proc/mounts that corresponds to
    ``device``, or the empty string if no mount point can be found.
    """
    with open("/proc/mounts", "r", encoding="utf8") as mounts:
        for line in mounts:
            fields = line.split(" ")
            if exists(fields[0]):
                if samefile(devpath, fields[0]):
                    return fields[1]
    return ""


class Manager:
    """
    Snapshot Manager high level interface.
    """

    plugins = []
    snapshot_sets = []
    by_name = {}
    by_uuid = {}

    @suspend_signals
    def __init__(self):
        self.plugins = []
        self.snapshot_sets = []
        self.by_name = {}
        self.by_uuid = {}
        check_boom_config()
        self._boot_cache = BootCache()
        plugin_classes = load_plugins()
        for plugin_class in plugin_classes:
            _log_debug("Loading plugin class '%s'", plugin_class.__name__)
            try:
                plugin = plugin_class(_log)
                self.plugins.append(plugin)
            except SnapmNotFoundError as err:
                _log_debug(
                    "Plugin dependencies missing: %s (%s), skipping.",
                    plugin_class.__name__,
                    err,
                )
            except SnapmPluginError as err:
                _log_error("Disabling plugin %s: %s", plugin_class.__name__, err)
        self.discover_snapshot_sets()

    def _find_and_verify_plugins(
        self, sources, size_policies, _requested_provider=None
    ):
        """
        Find snapshot provider plugins for each source in ``sources``
        and verify that a provider exists for each source present.

        :param sources: A list of source mount point or block device paths.
        :param size_policies: A dictionary mapping sources to size policies.
        :returns: A dictionary mapping sources to plugins.
        """
        # Initialise provider mapping.
        provider_map = {k: None for k in sources}

        # Find provider plugins for sources
        for source in sources:
            if not exists(source):
                _log_error("No such file or directory: %s", source)
                raise SnapmNotFoundError(f"Source path '{source}' does not exist")

            _log_debug(
                "Probing plugins for %s with size policy %s",
                source,
                size_policies[source],
            )

            if not ismount(source) and not S_ISBLK(stat(source).st_mode):
                raise SnapmPathError(
                    f"Path '{source}' is not a block device or mount point"
                )

            for plugin in self.plugins:
                if plugin.can_snapshot(source):
                    provider_map[source] = plugin

        # Verify each mount point has a provider plugin
        for source in provider_map:
            if provider_map[source] is None:
                raise SnapmNoProviderError(
                    f"Could not find snapshot provider for {source}"
                )
        return provider_map

    def _snapset_from_name_or_uuid(self, name=None, uuid=None):
        """
        Look a snapshot set up by ``name`` or ``uuid``. Returns a
        ``SnapshotSet`` correponding to either ``name`` or ``uuid`,
        or raises ``SnapmError`` on error.

        :param name: The name of the snapshot set to look up.
        :param uuid: The UUID of the snapshot set to look up.
        :returns: A ``SnapshotSet`` corresponding to the given name or UUID.
        :raises: ``SnapmNotFoundError`` is the name or UUID cannot be found.
                 ``SnapmInvalidIdentifierError`` if the name and UUID do not
                 match.
        """
        if name is not None:
            if name not in self.by_name:
                raise SnapmNotFoundError(f"Could not find snapshot set named {name}")
            snapset = self.by_name[name]
        elif uuid is not None:
            if uuid not in self.by_uuid:
                raise SnapmNotFoundError(
                    f"Could not find snapshot set with uuid {uuid}"
                )
            snapset = self.by_uuid[uuid]
        else:
            raise SnapmNotFoundError("A snapshot set name or UUID is required")

        if name and uuid:
            if self.by_name[name] != self.by_uuid[uuid]:
                raise SnapmInvalidIdentifierError(
                    f"Conflicting name and UUID: {str(uuid)} does not match '{name}'"
                )

        return snapset

    def _check_recursion(self, origins):
        """
        Verify that each entry in  ``origins`` corresponds to a device that is
        not a snapshot belonging to another snapshot set.

        :param origins: A list of origin devices to check.
        :raises: ``SnapmRecursionError`` if an origin device is a snapshot.
        """
        snapshot_devices = [
            snapshot.devpath
            for snapset in self.snapshot_sets
            for snapshot in snapset.snapshots
        ]
        for source, device in origins.items():
            if device in snapshot_devices:
                raise SnapmRecursionError(
                    "Snapshots of snapshots are not supported: "
                    f"{source} corresponds to snapshot device {device}"
                )

    def discover_snapshot_sets(self):
        """
        Discover snapshot sets by calling into each plugin to find
        individual snapshots and then aggregating them together into
        snapshot sets.

        Initialises the ``snapshot_sets``, ``by_name`` and ``by_uuid`` members
        with the discovered snapshot sets.
        """
        self.snapshot_sets = []
        self.by_name = {}
        self.by_uuid = {}
        snapshots = []
        snapset_names = set()
        self._boot_cache.refresh_cache()
        _log_debug("Discovering snapshot sets for %s plugins", len(self.plugins))
        for plugin in self.plugins:
            snapshots.extend(plugin.discover_snapshots())
        _log_debug("Discovered %s managed snapshots", len(snapshots))
        for snapshot in snapshots:
            snapset_names.add(snapshot.snapset_name)
        for snapset_name in snapset_names:
            set_snapshots = [
                snap for snap in snapshots if snap.snapset_name == snapset_name
            ]
            set_timestamp = set_snapshots[0].timestamp
            for snap in set_snapshots:
                if snap.timestamp != set_timestamp:
                    _log_warn(
                        "Snapshot set '%s' has inconsistent timestamps", snapset_name
                    )
                    continue

            snapset = SnapshotSet(snapset_name, set_timestamp, set_snapshots)

            # Associate snapset with boot entry if present
            if snapset.name in self._boot_cache.entry_cache:
                snapset.boot_entry = self._boot_cache.entry_cache[snapset.name]
            elif str(snapset.uuid) in self._boot_cache.entry_cache:
                snapset.boot_entry = self._boot_cache.entry_cache[str(snapset.uuid)]

            # Associate snapset with revert entry if present
            if snapset.name in self._boot_cache.revert_cache:
                snapset.revert_entry = self._boot_cache.revert_cache[snapset.name]
            elif str(snapset.uuid) in self._boot_cache.revert_cache:
                snapset.revert_entry = self._boot_cache.revert_cache[str(snapset.uuid)]

            self.snapshot_sets.append(snapset)
            self.by_name[snapset.name] = snapset
            self.by_uuid[snapset.uuid] = snapset
            for snapshot in snapset.snapshots:
                snapshot.snapshot_set = snapset

        self.snapshot_sets.sort(key=lambda ss: ss.name)
        _log_debug("Discovered %d snapshot sets", len(self.snapshot_sets))

    def find_snapshot_sets(self, selection=None):
        """
        Find snapshot sets matching selection criteria.

        :param selection: Selection criteria to apply.
        :returns: A list of matching ``SnapshotSet`` objects.
        """
        matches = []

        # Use null selection criteria if unspecified
        selection = selection if selection else Selection()

        selection.check_valid_selection(snapshot_set=True)

        _log_debug("Finding snapshot sets for %s", repr(selection))

        for snapset in self.snapshot_sets:
            if select_snapshot_set(selection, snapset):
                matches.append(snapset)

        _log_debug("Found %d snapshot sets", len(matches))
        return matches

    def find_snapshots(self, selection=None):
        """
        Find snapshots matching selection criteria.

        :param selection: Selection criteria to apply.
        :returns: A list of matching ``Snapshot`` objects.
        """
        matches = []

        # Use null selection criteria if unspecified
        selection = selection if selection else Selection()

        selection.check_valid_selection(snapshot=True, snapshot_set=True)

        _log_debug("Finding snapshots for %s", repr(selection))

        for snapset in self.snapshot_sets:
            for snapshot in snapset.snapshots:
                if select_snapshot(selection, snapshot):
                    matches.append(snapshot)

        _log_debug("Found %d snapshots", len(matches))
        return matches

    def _validate_snapset_name(self, name):
        """
        Validate a snapshot set name.

        Returns if ``name`` is a valid snapshot set name or raises an
        appropriate exception otherwise.

        :param name: The snapshot set name to validate.
        :raises: ``SnapmExistsError`` if the name is already in use, or
                 ``SnapmInvalidIdentifierError`` if the name fails validation.
        """
        if name in self.by_name:
            raise SnapmExistsError(f"Snapshot set named '{name}' already exists")
        for char in name:
            # Underscore is specifically disallowed in snapset names.
            if char == "_" or char not in SNAPM_VALID_NAME_CHARS:
                raise SnapmInvalidIdentifierError(
                    f"Snapshot set name cannot include '{char}'"
                )

    # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    @suspend_signals
    def create_snapshot_set(
        self, name, source_specs, default_size_policy=None, boot=False, revert=False
    ):
        """
        Create a snapshot set of the supplied mount points with the name
        ``name``.

        :param name: The name of the snapshot set.
        :param source_specs: A list of mount point and block device paths to
                             include in the set.
        :param default_size_policy: A default size policy to use for the set.
        :raises: ``SnapmExistsError`` if the name is already in use, or
                 ``SnapmInvalidIdentifierError`` if the name fails validation.
        """
        self._validate_snapset_name(name)

        # Parse size policies and normalise mount paths
        (sources, size_policies) = _parse_source_specs(
            source_specs, default_size_policy
        )

        # Initialise provider mapping.
        provider_map = self._find_and_verify_plugins(sources, size_policies, None)

        for provider in set(provider_map.values()):
            provider.start_transaction()

        timestamp = floor(time())
        origins = {}
        mounts = {}

        for source in provider_map:
            if S_ISBLK(stat(source).st_mode):
                mounts[source] = _find_mount_point_for_devpath(source)
                origins[source] = source
                mount = mounts[source]
                if mount in provider_map:
                    raise SnapmInvalidIdentifierError(
                        f"Duplicate snapshot source {source} already added to {name} as {mount}"
                    )
            else:
                mount = source
                origins[source] = provider_map[source].origin_from_mount_point(mount)

            try:
                provider_map[source].check_create_snapshot(
                    origins[source], name, timestamp, mount, size_policies[source]
                )
            except SnapmInvalidIdentifierError as err:
                _log_error(
                    "Error creating %s snapshot: %s", provider_map[source].name, err
                )
                raise SnapmInvalidIdentifierError(
                    f"Snapset name {name} too long"
                ) from err
            except SnapmNoSpaceError as err:
                _log_error(
                    "Error creating %s snapshot: %s", provider_map[source].name, err
                )
                raise SnapmNoSpaceError(
                    f"Insufficient free space for snapshot set {name}"
                ) from err

        self._check_recursion(origins)

        _suspend_journal()

        snapshots = []
        for source in provider_map:
            if S_ISBLK(stat(source).st_mode):
                mount = mounts[source]
            else:
                mount = source
            try:
                snapshots.append(
                    provider_map[source].create_snapshot(
                        origins[source], name, timestamp, mount, size_policies[source]
                    )
                )
            except SnapmError as err:
                _log_error("Error creating snapshot set member %s: %s", name, err)
                _resume_journal()
                for snapshot in snapshots:
                    snapshot.delete()
                raise SnapmPluginError(
                    f"Could not create all snapshots for set {name}"
                ) from err

        _resume_journal()

        for provider in set(provider_map.values()):
            provider.end_transaction()

        snapset = SnapshotSet(name, timestamp, snapshots)

        if boot or revert:
            _log_info(
                "Autoactivation required for bootable snapshot set '%s'", snapset.name
            )
            self._set_autoactivate(snapset, auto=True)
            snapset.activate()

        if boot:
            try:
                create_snapset_boot_entry(snapset)
                self._boot_cache.refresh_cache()
            except (OSError, ValueError) as err:
                _log_error("Failed to create snapshot set boot entry: %s", err)
                snapset.delete()
                raise SnapmCalloutError from err

        if revert:
            try:
                create_snapset_revert_entry(snapset)
                self._boot_cache.refresh_cache()
            except (OSError, ValueError) as err:
                _log_error("Failed to create snapshot set revert boot entry: %s", err)
                snapset.delete()
                raise SnapmCalloutError from err

        self.by_name[snapset.name] = snapset
        self.by_uuid[snapset.uuid] = snapset
        self.snapshot_sets.append(snapset)
        return snapset

    @suspend_signals
    def rename_snapshot_set(self, old_name, new_name):
        """
        Rename snapshot set ``old_name`` as ``new_name``.

        :param old_name: The name of the snapshot set to be renamed.
        :param new_name: The new name of the snapshot set.
        :raises: ``SnapmExistsError`` if the name is already in use, or
                 ``SnapmInvalidIdentifierError`` if the name fails validation.
        """
        if old_name not in self.by_name:
            raise SnapmNotFoundError(f"Cannot find snapshot set named {old_name}")

        self._validate_snapset_name(new_name)

        snapset = self.by_name[old_name]
        _check_snapset_status(snapset, "rename")

        # Remove references to old set
        self.by_name.pop(snapset.name)
        self.by_uuid.pop(snapset.uuid)
        self.snapshot_sets.remove(snapset)

        try:
            snapset.rename(new_name)
        except SnapmError as err:
            self.by_name[snapset.name] = snapset
            self.by_uuid[snapset.uuid] = snapset
            self.snapshot_sets.append(snapset)
            raise err

        self.by_name[snapset.name] = snapset
        self.by_uuid[snapset.uuid] = snapset
        self.snapshot_sets.append(snapset)
        return snapset

    @suspend_signals
    def delete_snapshot_sets(self, selection):
        """
        Remove snapshot sets matching selection criteria ``selection``.

        :param selection: Selection criteria for snapshot sets to remove.
        """
        sets = self.find_snapshot_sets(selection=selection)
        deleted = 0
        if not sets:
            raise SnapmNotFoundError(
                f"Could not find snapshot sets matching {selection}"
            )
        for snapset in sets:
            delete_snapset_boot_entry(snapset)
            delete_snapset_revert_entry(snapset)
            snapset.delete()
            self.snapshot_sets.remove(snapset)
            self.by_name.pop(snapset.name)
            self.by_uuid.pop(snapset.uuid)
            deleted += 1
        self._boot_cache.refresh_cache()
        return deleted

    # pylint: disable=too-many-branches
    @suspend_signals
    def resize_snapshot_set(
        self, source_specs, name=None, uuid=None, default_size_policy=None
    ):
        """
        Resize snapshot set named ``name`` or having UUID ``uuid``.

        Request to resize each snapshot included in ``source_specs``
        according to the given size policy, or apply ``default_size_policy``
        if set.

        :param name: The name of the snapshot set to resize.
        :param uuid: The UUID of the snapshot set to resize.
        :param source_specs: A list of mount points and optional size
                                  policies.
        :param default_size_policy: A default size policy to apply to the
                                    resize.
        """
        snapset = self._snapset_from_name_or_uuid(name=name, uuid=uuid)

        if source_specs:
            # Parse size policies and normalise mount paths
            (sources, size_policies) = _parse_source_specs(
                source_specs, default_size_policy
            )
        else:
            sources = [snapshot.source for snapshot in snapset.snapshots]
            size_policies = {source: default_size_policy for source in sources}

        snapset.resize(sources, size_policies)

    @suspend_signals
    def revert_snapshot_set(self, name=None, uuid=None):
        """
        Revert snapshot set named ``name`` or having UUID ``uuid``.

        Request to revert each snapshot origin within each snapshot set
        to the state at the time the snapshot was taken.

        :param name: The name of the snapshot set to revert.
        :param uuid: The UUID of the snapshot set to revert.
        """
        snapset = self._snapset_from_name_or_uuid(name=name, uuid=uuid)

        _check_revert_snapshot_set(snapset)

        # Snapshot boot entry becomes invalid as soon as revert is initiated.
        delete_snapset_boot_entry(snapset)

        snapset.revert()

        self._boot_cache.refresh_cache()
        return snapset

    def revert_snapshot_sets(self, selection):
        """
        Revert snapshot sets matching selection criteria ``selection``.

        Request to revert each snapshot origin within each snapshot set
        to the state at the time the snapshot was taken.

        :param selection: Selection criteria for snapshot sets to revert.
        """
        sets = self.find_snapshot_sets(selection=selection)
        reverted = 0
        if not sets:
            raise SnapmNotFoundError(
                f"Could not find snapshot sets matching {selection}"
            )
        for snapset in sets:
            self.revert_snapshot_set(name=snapset.name)
            reverted += 1
        return reverted

    @suspend_signals
    def activate_snapshot_sets(self, selection):
        """
        Activate snapshot sets matching selection criteria ``selection``.

        :param selection: Selection criteria for snapshot sets to activate.
        """
        sets = self.find_snapshot_sets(selection=selection)
        activated = 0
        if not sets:
            raise SnapmNotFoundError(
                f"Could not find snapshot sets matching {selection}"
            )
        for snapset in sets:
            _check_snapset_status(snapset, "activate")
            snapset.activate()
            activated += 1
        return activated

    @suspend_signals
    def deactivate_snapshot_sets(self, selection):
        """
        Deactivate snapshot sets matching selection criteria ``selection``.

        :param selection: Selection criteria for snapshot sets to deactivate.
        """
        sets = self.find_snapshot_sets(selection=selection)
        deactivated = 0
        if not sets:
            raise SnapmNotFoundError(
                f"Could not find snapshot sets matching {selection}"
            )
        for snapset in sets:
            _check_snapset_status(snapset, "deactivate")
            snapset.deactivate()
            deactivated += 1
        return deactivated

    def _set_autoactivate(self, snapset, auto=False):
        """
        Set autoactivation for ``snapset``.

        :param snapset: The ``SnapshotSet`` object to operate on.
        :param auto:  ``True`` to enable autoactivation or ``False`` otherwise.
        """
        _check_snapset_status(snapset, "set autoactivate status for")
        state = bool_to_yes_no(auto)
        _log_info("Setting autoactivation=%s for snapshot set %s", state, snapset.name)
        snapset.autoactivate = auto

    @suspend_signals
    def set_autoactivate(self, selection, auto=False):
        """
        Set autoactivation state for snapshot sets matching selection criteria
        ``selection``.

        :param selection: Selection criteria for snapshot sets to set
                          autoactivation.
        :param auto: ``True`` to enable autoactivation or ``False`` otherwise.
        """
        sets = self.find_snapshot_sets(selection=selection)
        changed = 0
        if not sets:
            raise SnapmNotFoundError(
                f"Could not find snapshot sets matching {selection}"
            )
        for snapset in sets:
            self._set_autoactivate(snapset, auto=auto)
            changed += 1
        return changed

    @suspend_signals
    def split_snapshot_set(self, name, new_name, source_specs):
        """
        Split the snapshot set named `name` into two snapshot sets, with
        `new_name` containing the listed `sources` and `name` containing all
        remaining snapshots.

        If `new_name` is `None` the listed snapshot are removed from `name`
        and permanently deleted. This operation cannot be undone.

        :param name: The name of the snapshot set to split
        :param new_name: The name of the newly split off snapshot set
        :param sources: The list of sources to include in the new snapshot set
        """
        if name not in self.by_name:
            raise SnapmNotFoundError(f"Cannot find snapshot set named {name}")

        if new_name:
            self._validate_snapset_name(new_name)

        op = f"{'split' if new_name else 'prune'}"

        _log_info("Attempting to %s snapshot set '%s'", op, name)

        snapset = self.by_name[name]
        _check_snapset_status(snapset, op)

        if not source_specs:
            raise SnapmArgumentError(
                f"SnapshotSet {op} requires at least one source argument"
            )

        # Parse size policies and normalise mount paths
        (sources, size_policies) = _parse_source_specs(source_specs, None)

        if any(size_policies[source] is not None for source in sources):
            err_sources = ", ".join(
                [source for source in source_specs if ":" in source]
            )
            raise SnapmArgumentError(
                f"SnapshotSet {op} does not support size policies: {err_sources}"
            )

        missing_sources = [
            source for source in sources if source not in snapset.sources
        ]

        if missing_sources:
            raise SnapmNotFoundError(
                f"SnapshotSet '{snapset.name}' does not contain {', '.join(missing_sources)}"
            )

        timestamp = snapset.timestamp

        split_snapshots = [
            snapshot for snapshot in snapset.snapshots if snapshot.source in sources
        ]

        keep_snapshots = [
            snapshot for snapshot in snapset.snapshots if snapshot.source not in sources
        ]

        if not keep_snapshots:
            op_str = "Splitting" if new_name else "Pruning"
            raise SnapmArgumentError(
                f"{op_str} {', '.join((snap.source for snap in split_snapshots))} "
                f"from {snapset.name} would leave {snapset.name} empty"
            )

        _log_debug(
            "Splitting snapshot%s %s from snapshot set '%s'",
            "s" if len(split_snapshots) > 1 else "",
            ", ".join((snap.source for snap in split_snapshots)),
            snapset.name,
        )

        self.by_name.pop(snapset.name)
        self.by_uuid.pop(snapset.uuid)
        self.snapshot_sets.remove(snapset)

        # Build new SnapshotSet with old name to start with, then use
        # SnapshotSet.rename() to propagate new_name to members.
        new_set = SnapshotSet(snapset.name, timestamp, split_snapshots)

        # Update original snapshot set and re-add to lookup tables
        orig_set = SnapshotSet(snapset.name, timestamp, keep_snapshots)
        self.by_name[orig_set.name] = orig_set
        self.by_uuid[orig_set.uuid] = orig_set
        self.snapshot_sets.append(orig_set)

        if new_name:
            # Attempt to rename members to new_name
            try:
                new_set.rename(new_name)
            except SnapmError as err:
                self.by_name[snapset.name] = snapset
                self.by_uuid[snapset.uuid] = snapset
                self.snapshot_sets.append(snapset)
                raise err
        else:
            try:
                # Delete the pruned members
                new_set.delete()
            except SnapmError as err:
                _log_error(
                    "Failed to clean up pruned snapshot set members: %s (%s)",
                    err,
                    ", ".join(snapshot.name for snapshot in new_set.snapshots),
                )
            return orig_set

        # Add new snapshot set to lookup tables
        self.by_name[new_set.name] = new_set
        self.by_uuid[new_set.uuid] = new_set
        self.snapshot_sets.append(new_set)

        return new_set

    @suspend_signals
    def create_snapshot_set_boot_entry(self, name=None, uuid=None):
        """
        Create a snapshot boot entry for the specified snapshot set.

        :param name: The name of the snapshot set.
        :param uuid: The UUID of the snapshot set.
        """
        snapset = self._snapset_from_name_or_uuid(name=name, uuid=uuid)

        snapset.autoactivate = True
        if not snapset.autoactivate:
            raise SnapmPluginError(
                "Could not enable autoactivation for all snapshots in snapshot "
                f"set {snapset.name}"
            )
        if snapset.boot_entry is not None:
            raise SnapmExistsError(
                f"Boot entry already associated with snapshot set {snapset.name}"
            )
        create_snapset_boot_entry(snapset)
        self._boot_cache.refresh_cache()

    @suspend_signals
    def create_snapshot_set_revert_entry(self, name=None, uuid=None):
        """
        Create a revert boot entry for the specified snapshot set.

        :param name: The name of the snapshot set.
        :param uuid: The UUID of the snapshot set.
        """
        snapset = self._snapset_from_name_or_uuid(name=name, uuid=uuid)

        if snapset.revert_entry is not None:
            raise SnapmExistsError(
                f"Revert entry already associated with snapshot set {snapset.name}"
            )
        create_snapset_revert_entry(snapset)
        self._boot_cache.refresh_cache()


__all__ = [
    "Manager",
]



# Copyright Red Hat
#
# tests/test_loader.py - Plugin loader tests
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: Apache-2.0
import unittest
import logging
import os
import re


import boom
from snapm.manager._loader import load_plugins
from snapm.manager.plugins import Plugin

from tests import BOOT_ROOT_TEST

log = logging.getLogger()

boom.set_boot_path(BOOT_ROOT_TEST)


class LoaderTestsSimple(unittest.TestCase):
    """
    Test manager interfaces with mock callouts
    """

    _old_path = None

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)
        bin_path = os.path.abspath("tests/bin")
        cur_path = os.environ["PATH"]
        self._old_path = cur_path
        os.environ["PATH"] = bin_path + os.pathsep + cur_path

    def tearDown(self):
        log.debug("Tearing down %s", self._testMethodName)
        os.environ["PATH"] = self._old_path

    def test_load_plugins(self):
        plugin_classes = load_plugins()
        self.assertEqual(len(plugin_classes), 3)

    def test_load_plugins_returns_plugins(self):
        plugin_classes = load_plugins()
        self.assertTrue(isinstance(plugin_classes, list))
        self.assertTrue(all(issubclass(c, Plugin) for c in plugin_classes))

    def test_load_plugins_have_name_and_version(self):
        rx = re.compile(r"\d+\.\d+\.\d+")
        for plugin_cls in load_plugins():
            self.assertTrue(isinstance(plugin_cls.name, str))
            self.assertTrue(isinstance(plugin_cls.version, str))
            self.assertTrue(rx.match(plugin_cls.version) is not None)



# Copyright 2022 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Error codes
"""
# isort: STDLIB
from enum import Enum, IntEnum


class Level(Enum):
    """
    Error levels.
    """

    ERROR = "E"
    WARNING = "W"
    INFO = "I"

    def __str__(self):
        return self.value


class PoolMaintenanceErrorCode(IntEnum):
    """
    Maintenance error codes for the pool.
    """

    NO_IPC_REQUESTS = 1
    NO_POOL_CHANGES = 2

    def __str__(self):
        return f"{Level.ERROR}M{str(self.value).zfill(3)}"

    @staticmethod
    def from_str(code_str):
        """
        Discover the code, if any, from the code string.

        :returns: the code if it finds a match, otherwise None
        :rtype: PoolMaintenanceErrorCode or NoneType
        """
        return next(
            (code for code in PoolMaintenanceErrorCode if code_str == str(code)), None
        )

    def explain(self):
        """
        Return an explanation of the error return code.
        """
        if self is PoolMaintenanceErrorCode.NO_IPC_REQUESTS:
            return (
                "The pool will return an error on any IPC request that could "
                "cause a change in the pool state, for example, a request to "
                "rename a filesystem. It will still be able to respond to "
                "purely informational requests."
            )

        if self is PoolMaintenanceErrorCode.NO_POOL_CHANGES:
            return (
                "The pool is unable to manage itself by reacting to events, "
                "such as devicemapper events, that might require it to take "
                "any maintenance operations."
            )

        assert False, "impossible error code reached"  # pragma: no cover

    def summarize(self):
        """
        Return a short summary of the return code.
        """
        if self is PoolMaintenanceErrorCode.NO_IPC_REQUESTS:
            return "Pool state changes not possible"

        if self is PoolMaintenanceErrorCode.NO_POOL_CHANGES:
            return "Pool maintenance operations not possible"

        assert False, "impossible error code reached"  # pragma: no cover


class PoolAllocSpaceErrorCode(IntEnum):
    """
    Code if the pool has run out of space to allocate.
    """

    NO_ALLOC_SPACE = 1

    def __str__(self):
        return f"{Level.WARNING}S{str(self.value).zfill(3)}"

    def explain(self):
        """
        Return an explanation of the return code.
        """
        if self is PoolAllocSpaceErrorCode.NO_ALLOC_SPACE:
            return (
                "Every device belonging to the pool has been fully allocated. "
                "To increase the allocable space, add additional data devices "
                "to the pool."
            )

        assert False, "impossible error code reached"  # pragma: no cover

    def summarize(self):
        """
        Return a short summary of the return code.
        """
        if self is PoolAllocSpaceErrorCode.NO_ALLOC_SPACE:
            return "All devices fully allocated"

        assert False, "impossible error code reached"  # pragma: no cover

    @staticmethod
    def from_str(code_str):
        """
        Discover the code, if any, from the code string.

        :returns: the code if it finds a match, otherwise None
        :rtype: PoolAllocSpaceErrorCode or NoneType
        """
        return next(
            (code for code in PoolAllocSpaceErrorCode if code_str == str(code)), None
        )


class PoolDeviceSizeChangeCode(IntEnum):
    """
    Codes for identifying for a pool if a device that belongs to the pool has
    been detected to have increased or reduced in size.
    """

    DEVICE_SIZE_INCREASED = 1
    DEVICE_SIZE_DECREASED = 2

    def __str__(self):
        if self is PoolDeviceSizeChangeCode.DEVICE_SIZE_INCREASED:
            return f"{Level.INFO}DS{str(self.value).zfill(3)}"

        if self is PoolDeviceSizeChangeCode.DEVICE_SIZE_DECREASED:
            return f"{Level.WARNING}DS{str(self.value).zfill(3)}"

        assert False, "impossible error code reached"  # pragma: no cover

    def explain(self):
        """
        Return an explanation of the return code.
        """
        if self is PoolDeviceSizeChangeCode.DEVICE_SIZE_INCREASED:
            return (
                "At least one device belonging to this pool appears to have "
                "increased in size."
            )

        if self is PoolDeviceSizeChangeCode.DEVICE_SIZE_DECREASED:
            return (
                "At least one device belonging to this pool appears to have "
                "decreased in size."
            )

        assert False, "impossible error code reached"  # pragma: no cover

    def summarize(self):
        """
        Return a short summary of the return code.
        """
        if self is PoolDeviceSizeChangeCode.DEVICE_SIZE_INCREASED:
            return "A device in this pool has increased in size."

        if self is PoolDeviceSizeChangeCode.DEVICE_SIZE_DECREASED:
            return "A device in this pool has decreased in size."

        assert False, "impossible error code reached"  # pragma: no cover

    @staticmethod
    def from_str(code_str):
        """
        Discover the code, if any, from the code string.

        :returns: the code if it finds a match, otherwise None
        :rtype: PoolAllocSpaceErrorCode or NoneType
        """
        return next(
            (code for code in PoolDeviceSizeChangeCode if code_str == str(code)), None
        )


class PoolErrorCode:
    """
    Summary class for all pool error codes.
    """

    CLASSES = [
        PoolMaintenanceErrorCode,
        PoolAllocSpaceErrorCode,
        PoolDeviceSizeChangeCode,
    ]

    @staticmethod
    def codes():
        """
        Return all pool error codes.
        """
        return [code for c in PoolErrorCode.CLASSES for code in list(c)]

    @staticmethod
    def error_from_str(error_code):
        """
        Obtain an error object from a distinguishing error string.

        :param str error_code:
        :returns: error object
        :raises: StopIteration if no match found
        """
        return next(
            (
                code
                for code in (c.from_str(error_code) for c in PoolErrorCode.CLASSES)
                if code is not None
            )
        )

    @staticmethod
    def explain(error_code):
        """
        Return explanation for error code, else None.

        :param str error_code:
        """
        return PoolErrorCode.error_from_str(error_code).explain()



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Error hierarchy for stratis cli.
"""


class StratisCliError(Exception):
    """
    Top-level stratis cli error.
    """


class StratisCliRuntimeError(StratisCliError):
    """
    Exception raised while an action is being performed and as a result of
    the requested action.
    """


class StratisCliGenerationError(StratisCliError):
    """
    Exception that occurs during generation of classes.
    """


class StratisCliEnvironmentError(StratisCliError):
    """
    Exception that occurs during processing of environment variables.
    """


class StratisCliStratisdVersionError(StratisCliRuntimeError):
    """
    Raised if stratisd version does not meet CLI version requirements.
    """

    def __init__(self, actual_version, minimum_version, maximum_version):
        """
        Initializer.
        :param tuple actual_version: stratisd's actual version
        :param tuple minimum_version: the minimum version required
        :param tuple maximum_version: the maximum version allowed
        """
        # pylint: disable=super-init-not-called
        self.actual_version = actual_version
        self.minimum_version = minimum_version
        self.maximum_version = maximum_version

    def __str__(self):
        fmt_str = (
            "stratisd version %s does not meet stratis version "
            "requirements; the version must be at least %s and no more "
            "than %s"
        )
        return fmt_str % (
            self.actual_version,
            self.minimum_version,
            self.maximum_version,
        )
